@@ -5,43 +5,39 @@ import datetime
 from sql import Literal
 from sql.aggregate import Count
 
-from trytond.model import ModelView, ModelSQL, MatchMixin, fields, \
-    sequence_ordered
-from trytond.pyson import Eval, If
+from trytond.i18n import gettext
+from trytond.model import (
+    ModelView, ModelSQL, MatchMixin, DeactivableMixin, sequence_ordered,
+    fields)
+from trytond.pyson import Eval, If, Bool
 from trytond.pool import Pool, PoolMeta
 from trytond.transaction import Transaction
+from trytond.tools import lstrip_wildcard
 
 from trytond.modules.product import price_digits
 
-__all__ = ['Template', 'Product', 'ProductSupplier', 'ProductSupplierPrice']
+from .exceptions import PurchaseUOMWarning
 
 
 class Template(metaclass=PoolMeta):
     __name__ = "product.template"
-    purchasable = fields.Boolean('Purchasable', states={
-            'readonly': ~Eval('active', True),
-            }, depends=['active'])
-    product_suppliers = fields.One2Many('purchase.product_supplier',
-        'product', 'Suppliers', states={
-            'readonly': ~Eval('active', True),
+    purchasable = fields.Boolean("Purchasable")
+    product_suppliers = fields.One2Many(
+        'purchase.product_supplier', 'template', "Suppliers",
+        states={
             'invisible': (~Eval('purchasable', False)
                 | ~Eval('context', {}).get('company')),
-            }, depends=['active', 'purchasable'])
+            },
+        domain=[
+            If(~Eval('active'), ('active', '=', False), ()),
+            ],
+        depends=['purchasable', 'active'])
     purchase_uom = fields.Many2One('product.uom', 'Purchase UOM', states={
-            'readonly': ~Eval('active'),
             'invisible': ~Eval('purchasable'),
             'required': Eval('purchasable', False),
             },
         domain=[('category', '=', Eval('default_uom_category'))],
-        depends=['active', 'purchasable', 'default_uom_category'])
-
-    @classmethod
-    def __setup__(cls):
-        super(Template, cls).__setup__()
-        cls._error_messages.update({
-                'change_purchase_uom': ('Purchase prices are based '
-                    'on the purchase uom.'),
-                })
+        depends=['purchasable', 'default_uom_category'])
 
     @fields.depends('default_uom', 'purchase_uom', 'purchasable')
     def on_change_default_uom(self):
@@ -63,8 +59,18 @@ class Template(metaclass=PoolMeta):
                     'invisible': ~Eval('purchasable'),
                     })]
 
+    def product_suppliers_used(self, **pattern):
+        # Skip rules to test pattern on all records
+        with Transaction().set_user(0):
+            template = self.__class__(self)
+        for product_supplier in template.product_suppliers:
+            if product_supplier.match(pattern):
+                yield product_supplier
+
     @classmethod
     def write(cls, *args):
+        pool = Pool()
+        Warning = pool.get('res.user.warning')
         actions = iter(args)
         for templates, values in zip(actions, actions):
             if not values.get("purchase_uom"):
@@ -77,15 +83,52 @@ class Template(metaclass=PoolMeta):
                 for product in template.products:
                     if not product.product_suppliers:
                         continue
-                    cls.raise_user_warning(
-                            '%s@product_template' % template.id,
-                            'change_purchase_uom')
+                    name = '%s@product_template' % template.id
+                    if Warning.check(name):
+                        raise PurchaseUOMWarning(
+                            name, gettext('purchase.msg_change_purchase_uom'))
         super(Template, cls).write(*args)
 
+    @classmethod
+    def copy(cls, templates, default=None):
+        pool = Pool()
+        ProductSupplier = pool.get('purchase.product_supplier')
+        if default is None:
+            default = {}
+        else:
+            default = default.copy()
+
+        copy_suppliers = 'product_suppliers' not in default
+        default.setdefault('product_suppliers', None)
+        new_templates = super().copy(templates, default)
+        if copy_suppliers:
+            old2new = {}
+            to_copy = []
+            for template, new_template in zip(templates, new_templates):
+                to_copy.extend(
+                    ps for ps in template.product_suppliers if not ps.product)
+                old2new[template.id] = new_template.id
+            if to_copy:
+                ProductSupplier.copy(to_copy, {
+                        'template': lambda d: old2new[d['template']],
+                        })
+        return new_templates
+
 
 class Product(metaclass=PoolMeta):
     __name__ = 'product.product'
 
+    product_suppliers = fields.One2Many(
+        'purchase.product_supplier', 'product', "Suppliers",
+        domain=[
+            ('template', '=', Eval('template')),
+            If(~Eval('active'), ('active', '=', False), ()),
+            ],
+        states={
+            'invisible': (~Eval('purchasable', False)
+                | ~Eval('context', {}).get('company')),
+            },
+        depends=['template', 'purchasable', 'active'])
     purchase_price_uom = fields.Function(fields.Numeric(
             "Purchase Price", digits=price_digits), 'get_purchase_price_uom')
 
@@ -94,6 +137,19 @@ class Product(metaclass=PoolMeta):
         quantity = Transaction().context.get('quantity') or 0
         return cls.get_purchase_price(products, quantity=quantity)
 
+    def product_suppliers_used(self, **pattern):
+        # Skip rules to test pattern on all records
+        with Transaction().set_user(0):
+            product = self.__class__(self)
+        for product_supplier in product.product_suppliers:
+            if product_supplier.match(pattern):
+                yield product_supplier
+        pattern['product'] = None
+        yield from self.template.product_suppliers_used(**pattern)
+
+    def _get_purchase_unit_price(self, quantity=0):
+        return self.get_multivalue('cost_price')
+
     @classmethod
     def get_purchase_price(cls, products, quantity=0):
         '''
@@ -128,7 +184,8 @@ class Product(metaclass=PoolMeta):
         user = User(Transaction().user)
 
         for product in products:
-            prices[product.id] = product.cost_price
+            prices[product.id] = product._get_purchase_unit_price(
+                quantity=quantity)
             default_uom = product.default_uom
             default_currency = (user.company.currency if user.company
                 else None)
@@ -137,21 +194,18 @@ class Product(metaclass=PoolMeta):
             else:
                 product_uom = uom
             pattern = ProductSupplier.get_pattern()
-            product_suppliers = product.product_suppliers
-            if context.get('product_supplier'):
-                for product_supplier in product.product_suppliers:
-                    if product_supplier.id == context['product_supplier']:
-                        product_suppliers = [product_supplier]
-                        break
-            for product_supplier in product_suppliers:
-                if product_supplier.match(pattern):
-                    pattern = ProductSupplierPrice.get_pattern()
-                    for price in product_supplier.prices:
-                        if price.match(quantity, product_uom, pattern):
-                            prices[product.id] = price.unit_price
-                            default_uom = product_supplier.uom
-                            default_currency = product_supplier.currency
-                    break
+            product_suppliers = product.product_suppliers_used(**pattern)
+            try:
+                product_supplier = next(product_suppliers)
+            except StopIteration:
+                pass
+            else:
+                pattern = ProductSupplierPrice.get_pattern()
+                for price in product_supplier.prices:
+                    if price.match(quantity, product_uom, pattern):
+                        prices[product.id] = price.unit_price
+                        default_uom = product_supplier.uom
+                        default_currency = product_supplier.currency
             prices[product.id] = Uom.compute_price(
                 default_uom, prices[product.id], product_uom)
             if currency and default_currency:
@@ -161,27 +215,92 @@ class Product(metaclass=PoolMeta):
                         prices[product.id], currency, round=False)
         return prices
 
+    @classmethod
+    def copy(cls, products, default=None):
+        pool = Pool()
+        ProductSupplier = pool.get('purchase.product_supplier')
+        if default is None:
+            default = {}
+        else:
+            default = default.copy()
 
-class ProductSupplier(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
+        copy_suppliers = 'product_suppliers' not in default
+        if 'template' in default:
+            default.setdefault('product_suppliers', None)
+        new_products = super().copy(products, default)
+        if 'template' in default and copy_suppliers:
+            template2new = {}
+            product2new = {}
+            to_copy = []
+            for product, new_product in zip(products, new_products):
+                if product.product_suppliers:
+                    to_copy.extend(product.product_suppliers)
+                    template2new[product.template.id] = new_product.template.id
+                    product2new[product.id] = new_product.id
+            if to_copy:
+                ProductSupplier.copy(to_copy, {
+                        'product': lambda d: product2new[d['product']],
+                        'template': lambda d: template2new[d['template']],
+                        })
+        return new_products
+
+
+class ProductSupplier(
+        sequence_ordered(), DeactivableMixin, ModelSQL, ModelView, MatchMixin):
     'Product Supplier'
     __name__ = 'purchase.product_supplier'
-    product = fields.Many2One('product.template', 'Product', required=True,
-            ondelete='CASCADE', select=True)
-    party = fields.Many2One('party.party', 'Supplier', required=True,
-        ondelete='CASCADE', select=True)
+    template = fields.Many2One(
+        'product.template', "Product",
+        required=True, ondelete='CASCADE', select=True,
+        domain=[
+            If(Bool(Eval('product')),
+                ('products', '=', Eval('product')),
+                ()),
+            If(Eval('active'), ('active', '=', True), ()),
+            ],
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=['product', 'active', 'company'])
+    product = fields.Many2One(
+        'product.product', "Variant", select=True,
+        domain=[
+            If(Bool(Eval('template')),
+                ('template', '=', Eval('template')),
+                ()),
+            If(Eval('active'), ('active', '=', True), ()),
+            ],
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=['template', 'active', 'company'])
+    party = fields.Many2One(
+        'party.party', 'Supplier', required=True, ondelete='CASCADE',
+        select=True,
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=['company'])
     name = fields.Char('Name', size=None, translate=True, select=True)
     code = fields.Char('Code', size=None, select=True)
     prices = fields.One2Many('purchase.product_supplier.price',
             'product_supplier', 'Prices')
-    company = fields.Many2One('company.company', 'Company', required=True,
-        ondelete='CASCADE', select=True,
-        domain=[
-            ('id', If(Eval('context', {}).contains('company'), '=', '!='),
-                Eval('context', {}).get('company', -1)),
-            ])
-    lead_time = fields.TimeDelta('Lead Time')
+    company = fields.Many2One(
+        'company.company', "Company",
+        required=True, ondelete='CASCADE', select=True)
+    lead_time = fields.TimeDelta('Lead Time',
+        help="The time from confirming the purchase order to receiving the "
+        "products.\n"
+        "If empty the lead time of the supplier is used.")
     currency = fields.Many2One('currency.currency', 'Currency', required=True,
         ondelete='RESTRICT')
+    uom = fields.Function(
+        fields.Many2One('product.uom', "UOM"), 'on_change_with_uom')
+
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls.__access__.update(['product', 'template'])
 
     @classmethod
     def __register__(cls, module_name):
@@ -190,13 +309,18 @@ class ProductSupplier(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
         table = cls.__table_handler__(module_name)
         sql_table = cls.__table__()
 
+        # Migration from 5.0: add product/template
+        if (table.column_exist('product')
+                and not table.column_exist('template')):
+            table.column_rename('product', 'template')
+
         super(ProductSupplier, cls).__register__(module_name)
 
         # Migration from 3.8: change delivery_time inte timedelta lead_time
         if table.column_exist('delivery_time'):
             cursor.execute(*sql_table.select(
                     sql_table.id, sql_table.delivery_time))
-            for id_, delivery_time in cursor.fetchall():
+            for id_, delivery_time in cursor:
                 if delivery_time is None:
                     continue
                 lead_time = datetime.timedelta(days=delivery_time)
@@ -217,6 +341,12 @@ class ProductSupplier(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
             company = Company(Transaction().context['company'])
             return company.currency.id
 
+    @fields.depends(
+        'product', '_parent_product.template')
+    def on_change_product(self):
+        if self.product:
+            self.template = self.product.template
+
     @fields.depends('party')
     def on_change_party(self):
         cursor = Transaction().connection.cursor()
@@ -232,11 +362,15 @@ class ProductSupplier(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
                 self.currency, = row
 
     def get_rec_name(self, name):
-        if self.code:
-            name = self.name or self.product.name
-            return '[' + self.code + '] ' + name
+        if self.name:
+            name = self.name
+        elif self.product:
+            name = self.product.name
         else:
-            return self.name or self.product.rec_name
+            name = self.template.name
+        if self.code:
+            name = '[' + self.code + '] ' + name
+        return name
 
     @classmethod
     def search_rec_name(cls, name, clause):
@@ -244,17 +378,39 @@ class ProductSupplier(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
             bool_op = 'AND'
         else:
             bool_op = 'OR'
+        code_value = clause[2]
+        if clause[1].endswith('like'):
+            code_value = lstrip_wildcard(clause[2])
         domain = [bool_op,
+            ('template',) + tuple(clause[1:]),
             ('product',) + tuple(clause[1:]),
             ('party',) + tuple(clause[1:]),
-            ('code',) + tuple(clause[1:]),
+            ('code', clause[1], code_value) + tuple(clause[3:]),
             ('name',) + tuple(clause[1:]),
             ]
         return domain
 
+    @fields.depends(
+        'product', '_parent_product.purchase_uom',
+        'template', '_parent_template.purchase_uom')
+    def on_change_with_uom(self, name=None):
+        if self.product:
+            if self.product.purchase_uom:
+                return self.product.purchase_uom.id
+        elif self.template:
+            if self.template.purchase_uom:
+                return self.template.purchase_uom.id
+
     @property
-    def uom(self):
-        return self.product.purchase_uom
+    def lead_time_used(self):
+        # Use getattr because it can be called with an unsaved instance
+        lead_time = getattr(self, 'lead_time', None)
+        party = getattr(self, 'party', None)
+        if lead_time is None and party:
+            company = getattr(self, 'company', None)
+            lead_time = party.get_multivalue(
+                'supplier_lead_time', company=company.id if company else None)
+        return lead_time
 
     def compute_supply_date(self, date=None):
         '''
@@ -264,9 +420,9 @@ class ProductSupplier(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
 
         if not date:
             date = Date.today()
-        if self.lead_time is None:
+        if self.lead_time_used is None:
             return datetime.date.max
-        return date + self.lead_time
+        return date + self.lead_time_used
 
     def compute_purchase_date(self, date):
         '''
@@ -274,16 +430,17 @@ class ProductSupplier(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
         '''
         Date = Pool().get('ir.date')
 
-        if self.lead_time is None:
+        if self.lead_time_used is None or date is None:
             return Date.today()
-        return date - self.lead_time
+        return date - self.lead_time_used
 
     @staticmethod
     def get_pattern():
         context = Transaction().context
-        return {
-            'party': context.get('supplier'),
-            }
+        pattern = {'party': context.get('supplier')}
+        if 'product_supplier' in context:
+            pattern['id'] = context['product_supplier']
+        return pattern
 
 
 class ProductSupplierPrice(
@@ -296,10 +453,22 @@ class ProductSupplierPrice(
         "Quantity",
         required=True,
         domain=[('quantity', '>=', 0)],
-        help='Minimal quantity')
+        help='Minimal quantity.')
     unit_price = fields.Numeric('Unit Price', required=True,
         digits=price_digits)
 
+    uom = fields.Function(fields.Many2One('product.uom', 'UOM',
+        help="The unit in which the quantity is specified."),
+        'on_change_with_uom')
+    currency = fields.Function(
+        fields.Many2One('currency.currency', 'Currency'),
+        'on_change_with_currency')
+
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls.__access__.add('product_supplier')
+
     @classmethod
     def __register__(cls, module_name):
         cursor = Transaction().connection.cursor()
@@ -319,6 +488,16 @@ class ProductSupplierPrice(
     def default_quantity():
         return 0.0
 
+    @fields.depends('product_supplier', '_parent_product_supplier.product')
+    def on_change_with_uom(self, name=None):
+        if self.product_supplier and self.product_supplier.uom:
+            return self.product_supplier.uom.id
+
+    @fields.depends('product_supplier', '_parent_product_supplier.currency')
+    def on_change_with_currency(self, name=None):
+        if self.product_supplier and self.product_supplier.currency:
+            return self.product_supplier.currency.id
+
     @staticmethod
     def get_pattern():
         return {}
