@@ -10,11 +10,13 @@ from trytond.tests.test_tryton import doctest_teardown, doctest_checker
 from trytond.transaction import Transaction
 from trytond.pool import Pool
 
-from trytond.modules.company.tests import create_company, set_company
+from trytond.modules.company.tests import (
+    create_company, set_company, PartyCompanyCheckEraseMixin, CompanyTestMixin)
 from trytond.modules.account.tests import create_chart
 
 
-class PurchaseTestCase(ModuleTestCase):
+class PurchaseTestCase(
+        PartyCompanyCheckEraseMixin, CompanyTestMixin, ModuleTestCase):
     'Test Purchase module'
     module = 'purchase'
 
@@ -34,11 +36,11 @@ class PurchaseTestCase(ModuleTestCase):
             create_chart(company)
 
             receivable, = Account.search([
-                    ('kind', '=', 'receivable'),
+                    ('type.receivable', '=', True),
                     ('company', '=', company.id),
                     ])
             payable, = Account.search([
-                    ('kind', '=', 'payable'),
+                    ('type.payable', '=', True),
                     ('company', '=', company.id),
                     ])
 
@@ -62,7 +64,7 @@ class PurchaseTestCase(ModuleTestCase):
                         'account_payable': payable.id,
                         }])
             product_supplier, = ProductSupplier.create([{
-                        'product': template.id,
+                        'template': template.id,
                         'party': supplier.id,
                         'prices': [('create', [{
                                         'sequence': 1,
@@ -122,4 +124,14 @@ def suite():
             tearDown=doctest_teardown, encoding='utf-8',
             optionflags=doctest.REPORT_ONLY_FIRST_FAILURE,
             checker=doctest_checker))
+    suite.addTests(doctest.DocFileSuite(
+            'scenario_purchase_return_wizard.rst',
+            tearDown=doctest_teardown, encoding='utf-8',
+            optionflags=doctest.REPORT_ONLY_FIRST_FAILURE,
+            checker=doctest_checker))
+    suite.addTests(doctest.DocFileSuite(
+            'scenario_purchase_copy_product_suppliers.rst',
+            tearDown=doctest_teardown, encoding='utf-8',
+            optionflags=doctest.REPORT_ONLY_FIRST_FAILURE,
+            checker=doctest_checker))
     return suite
